DAT fileName  byte "AS5055Demo3Wire130730f", 0
CON
{{ ****** Public Notes ******
  
  Written by: Duane Degn
  For: Kirk Fraser
  March 24, 2012

  See comments in child object.

  This demo only uses the 4-wire daisy-chain
  capability of the AS5055Spi120324a object.

  In daisy-chain mode the device at the end
  of the chain is considered device #0.

  Presently this object outputs data to a
  terminal window only when there is a change
  in the value to display.
                                  
  The display is formatted on the terminal window.
  With the new value overwriting the previous value.

  If you'd like to preserve a record of the values
  in the terminal window, change the value of
  "displayMode" in the first line of the Setup
  method to "SCROLL_DISPLAY".
  
  This will cause the values to scroll down
  the terminal window leaving a record of the readings.

  
}}
{ ****** Private Notes ****** 
  Used servo tester as starting point.
  120316a Start writing software to read
  AS5055 magnetic encoders.
  17a Modify SPI driver to read 16 bits.
  Start with three wire read only mode.
  17b SPI object returns 32 bits sometimes.
  Why?
  18a Try again with MOSI pulled high.
  18a Appears to work a lot better.
  18b Clear the temp variable in the
  child object so return values are 16 bits
  not 32.
  18c Still 32 bits.  I had cleared the wrong
  variable.
  18d the parity bit is always zero.  I think
  I'm reading in one too many bits.
  18e Change the way child reads data in.
  18e Works much better than previous versions.
  18f Change output to 7-segment display.
  18g Change to expect twelve bits.  Instead of
  14 bits.
  18g Works really well with three wire SPI
  read only.
  LED display still needs some formatting help.
  18h Try to add second display mode to output
  raw values.
  Add a timer to change display modes.
  If output is between 40 and 50 degrees for two
  seconds and then moves to between 120 and 130 degrees
  within half a second and remains there for two seconds,
  the output will change to raw data mode.
  18k Add second PST object for 7-segment display.
  18k Read out jumps around with last digit.
  20a Try to get a more stable reading by averaging
  values.
  20a Works well except when near transition from
  4095 to 0.  The average near this transition
  can produce strange results.
  20b Try to come up with a way to smooth out
  transition.
  20c Seems to be working.
  21a Start adding daisy-chain capability.
  21b Simplify program.
  23b Daisy-chaining works!
  23d Add another buffer for averaging fast moves.
  23d didn't work correctly.  The displays froze up.
  23e Still didn't work.  AverageM equaled zero.
  23f Abandon moving average stuff.
  Work on cleaning up code.
  I don't think the second encoder averaging
  works yet.
  23g Strip code down to bare minimum for now.
  22bb works with two averaging buffers.
  23i Revert to 23e try to add moving average back in.
  23i Works.
  24a Clean up code, comment out 7-segment display code.
  AS5055Demo3Wire130728a  Start moving filter to child object.
  130728a Change constants to gold standard format.
  130728B Use 3-wire mode with no interrupt.
  29a Seems to work okay.
  30b Still works.
  30c Try using continuous mode.
  30c Back to single read mode. Works.
  30d Try continuous mode again.
  30f Continuous mode kind of works but every other read is a zero.
  
  Things to add.
  Check parity bit.
 
}
CON

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                    

  CLK_FREQ = ((_clkmode - xtal1) >> 6) * _xinfreq
  MS_001   = CLK_FREQ / 1_000
  US_001   = CLK_FREQ / 1_000_000

  TWELVE_BITS = %1111_1111_1111
  TWELVE_BITS_PLUS_ONE = TWELVE_BITS + 1

  '' Set these pins to correspond with how the
  '' devices are connected to the Propeller chip.
  INT_PIN = -1 '8   ' input active low
  MOSI_PIN = -1 '9  ' output
  MISO_PIN = 8 '10 ' input
  SCK_PIN = 10 '11  ' output
  SS_PIN = 12   ' output active low

  '' Baud rate used to communicate with the PC terminal window.
  PST_BAUD = 115_200 '57600

  ' Constants used with the small 7-segment displays
  {SEVEN_SEGMENT_BAUD = 9600 
  DEBUG_SEVEN_SEGMENT_TX_PIN_0 = 14
  DEBUG_SEVEN_SEGMENT_TX_PIN_1 = 15
  
  SPECIAL_CHARACTER = $77
  DECIMAL_POINT_0 = %1
  DECIMAL_POINT_1 = %10
  DECIMAL_POINT_2 = %100
  DECIMAL_POINT_Off = 0  }

  '' The three constants (SAMPLES_TO_AVERAGE_STATIC, SAMPLES_TO_AVERAGE_MOVING and
  '' MOVING_TRIGGER) control the software filters.
  ''
  '' Small sample values will give faster position updates at the expense
  '' of less stable readings.
  ''
  '' MOVING_TRIGGER is how large a change from one reading to the next
  '' is needed to trigger the smaller sample set used to compute the average.
  SAMPLES_TO_AVERAGE_STATIC = 64 '128  ' should be a power of two
  SAMPLES_TO_AVERAGE_MOVING = 16  ' should be a power of two
  MOVING_TRIGGER = 8            ' difference between two concecutive readings to trigger using small buffer to average
  ROLLOVER_ZONE = TWELVE_BITS_PLUS_ONE >> 2 '1000
  ROLLOVER_ZONE_LOW = ROLLOVER_ZONE
  ROLLOVER_ZONE_HIGH = TWELVE_BITS - ROLLOVER_ZONE
  
  '' Set the constant "CHAINED_DEVICES" to the number of devices that
  '' are to be monitored in daisy-chain mode. 
  CHAINED_DEVICES = 1
  BUFFER_SIZE_STATIC = SAMPLES_TO_AVERAGE_STATIC * CHAINED_DEVICES
  BUFFER_SIZE_MOVING = SAMPLES_TO_AVERAGE_MOVING * CHAINED_DEVICES
  
  COLUMN_WIDTH = 12   ' Terminal Window Formatting Constants
  ROW_WIDTH = 4
  DEVICES_PER_ROW = 4
  DISPLAY_START_Y = 10
  DISPLAY_START_X = 4
  TITLE_TO_READING_ROWS = 1
  DISPLAY_DIGITS = 4
  READ_DELAY = 200 '100
  LIMIT_ERROR_MESSAGES = 10
  
  ' movingState enumeration
  #0, NOT_MOVING_UPDATED_BUFFER, MOVING, NOT_MOVING_FILLING_BUFFER
    
  ' rolloverState enumeration
  #0, OUT_OF_ZONE_EMPTY_BUFFER, IN_ZONE_FILLING_BUFFER, IN_ZONE_USE_BUFFER, OUT_OF_ZONE_EMPTYING_BUFFER

  ' localMovingFlag enumeration
  #0, NOTMOVING_THIS_LOOP, MOVING_THIS_LOOP

  ' displayMode enumeration
  #0, SCROLL_DISPLAY, STATIONARY_DISPLAY

  ' WIRE_MODE enumeration
  #0, THREE_WIRE_MODE, FOUR_WIRE_MODE

  ' INTERRUPT_MODE enumeration
  #0, NO_INTERRUPT, YES_INTERRUPT

  INTERRUPT_MODE = NO_INTERRUPT
  WIRE_MODE = THREE_WIRE_MODE
  
VAR

  long spiCog, encoderAngle[CHAINED_DEVICES], readingsCounter
  long encoderAngle7[CHAINED_DEVICES], previousAngle[CHAINED_DEVICES]
  long angleDegrees, pseudoRealMultiplier, fullCircleDegrees
  'long transitionMode, from1stLow[2], from1stHigh[2], from2ndLow[2], from2ndHigh[2] 
  long transitionInterval, transitionTime, holdPositionInterval
  
  ' Variables duplicated in PASM of child object.
  long bufferHead
  long bufferHeadM
  ' Variables used by child object. Keep the varaiables below together and in order.
  long samplesToAverageStatic, samplesToAverageMoving, movingThreshold
  long bufferSizeStatic, bufferSizeMoving, wireMode, interruptMode
  
  long bufferTotal[CHAINED_DEVICES], bufferTotalMov[CHAINED_DEVICES]
  long average[CHAINED_DEVICES], averageMov[CHAINED_DEVICES]
  long rolloverTotal[CHAINED_DEVICES], rolloverTotalMov[CHAINED_DEVICES]
  long rolloverCount[CHAINED_DEVICES], rolloverCountMov[CHAINED_DEVICES]
  long notMovingCount[CHAINED_DEVICES]
 
  long averagingBuffer[BUFFER_SIZE_STATIC], averagingBufferMov[BUFFER_SIZE_MOVING]
  long rolloverBuffer[BUFFER_SIZE_STATIC], rolloverBufferMov[BUFFER_SIZE_MOVING]
  ' Keep the varaiables above together and in order.
  byte movingState[CHAINED_DEVICES]
  byte rolloverState[CHAINED_DEVICES], rolloverStateMov[CHAINED_DEVICES]
  byte displayX[CHAINED_DEVICES], displayY[CHAINED_DEVICES]
  byte titleX[CHAINED_DEVICES], titleY[CHAINED_DEVICES]
  byte counterLocX, counterLocY
  byte colHeadingSize, dataDisplayOffset, displayMode
  byte chainedDevices, finalChainIndex, errorFlag
    
OBJ

  Debug : "Parallax Serial Terminal"
  'SevenSegment[2] : "Parallax Serial Terminal"  
  Spi : "AS5055Spi130730e"
  
PUB Setup | localIndex
'' First method executed by the Propeller.
'' This method initializes variables and
'' starts drivers.  

  displayMode := SCROLL_DISPLAY 'STATIONARY_DISPLAY  ' 
  
  samplesToAverageStatic := SAMPLES_TO_AVERAGE_STATIC
  samplesToAverageMoving := SAMPLES_TO_AVERAGE_MOVING
  movingThreshold := MOVING_TRIGGER
  bufferSizeStatic := BUFFER_SIZE_STATIC
  bufferSizeMoving := BUFFER_SIZE_MOVING
  wireMode := WIRE_MODE
  interruptMode := INTERRUPT_MODE
  colHeadingSize := strsize(@columnHeading)
  dataDisplayOffset := ((colHeadingSize + 1) - DISPLAY_DIGITS) / 2
  if WIRE_MODE == THREE_WIRE_MODE
    chainedDevices := 1
  else
    chainedDevices := CHAINED_DEVICES
    
  finalChainIndex := chainedDevices - 1
    
  repeat localIndex from 0 to finalChainIndex
    rolloverState[localIndex] := OUT_OF_ZONE_EMPTY_BUFFER
    rolloverStateMov[localIndex] := OUT_OF_ZONE_EMPTY_BUFFER
    movingState[localIndex] := MOVING
    titleY[localIndex] := DISPLAY_START_Y + ((localIndex / DEVICES_PER_ROW) * ROW_WIDTH) 
    displayY[localIndex] := titleY[localIndex] + TITLE_TO_READING_ROWS
    titleX[localIndex] := (localIndex // DEVICES_PER_ROW) * COLUMN_WIDTH
    displayX[localIndex] := titleX[localIndex] + dataDisplayOffset 
  counterLocX := displayX[finalChainIndex] + 22
  counterLocY := displayY[finalChainIndex]
       
  {SevenSegment[0].StartRxTx(-1, DEBUG_SEVEN_SEGMENT_TX_PIN_0, 0, SEVEN_SEGMENT_BAUD)
  SevenSegment[1].StartRxTx(-1, DEBUG_SEVEN_SEGMENT_TX_PIN_1, 0, SEVEN_SEGMENT_BAUD)
  
  SevenSegment[0].Char("z")
  SevenSegment[0].Char(0)   ' Turn up brightness?
  SevenSegment[1].Char("z")
  SevenSegment[1].Char(0)   ' Turn up brightness?
  }
  
  Debug.Start(PST_BAUD)
    
    
  waitcnt(clkfreq / 20 + cnt)
  spiCog := Spi.Start(MOSI_PIN, MISO_PIN, SCK_PIN, SS_PIN, READ_DELAY * MS_001, {
                    } @samplesToAverageStatic, @bufferTotal) - 1
  
  waitcnt(clkfreq * 3 + cnt) ' Time to open terminal window.                          
  DisplayHeading
  if WIRE_MODE == FOUR_WIRE_MODE
    Spi.Reset(chainedDevices)
  'Debug.str(string(13, "Before ReadChain call"))  
  Spi.ReadContinuous(@encoderAngle, chainedDevices)   
  longfill(@previousAngle, -1, chainedDevices)  
  

  'Debug.str(string(13, "After ReadChain call"))
  {if INTERRUPT_MODE == YES_INTERRUPT  
    WithInterrupt
  else
    NoInterrupt   }

  MainLoop
    
PUB MainLoop | localIndex
                                                                                                        
  repeat
    repeat localIndex from 0 to finalChainIndex
      if encoderAngle[localIndex] <> previousAngle[localIndex]
      ' Comment out above line to display encoder value in each loop.
        if displayMode == SCROLL_DISPLAY 
          Debug.str(string(13, " encoder["))
          Debug.dec(localIndex)
          Debug.str(string("] = "))
          Debug.dec(encoderAngle[localIndex])
          Debug.PositionX(counterLocX)
          Debug.Dec(readingsCounter)
          Debug.str(string("   "))    
        else        
          DisplayTitles
          DisplayData(localIndex)
        previousAngle[localIndex] := encoderAngle[localIndex]
        
PUB NoInterrupt

  'Debug.str(string(13, "NoInterrupt method"))                                                                                                         
  repeat    
    Pause(READ_DELAY)
    ReadSensor
       
PUB WithInterrupt | pinState
                                                
  'Debug.str(string(13, "WithInterrupt method"))
  repeat
    pinState := ina[INT_PIN]
    if pinState == 0
      ReadSensor
        
PUB ReadSensor | localIndex
    
  'Debug.str(string(13, "ReadSensor method"))
  Spi.ReadChain(@encoderAngle, chainedDevices)
        
  if ++bufferHead == SAMPLES_TO_AVERAGE_STATIC
    bufferHead := 0  
  if ++bufferHeadM == SAMPLES_TO_AVERAGE_MOVING
    bufferHeadM := 0
 
  repeat localIndex from 0 to finalChainIndex
    errorFlag += ProcessAngle(localIndex)
  if (errorFlag > 0 and WIRE_MODE == FOUR_WIRE_MODE) or errorFlag == 1
    Debug.str(string(13, "                              ****** Error ****** "))
    if WIRE_MODE == FOUR_WIRE_MODE 
      Debug.str(string(13, " Reseting Devices "))
      Spi.Reset(chainedDevices)                      ' I'm not sure if this will clear the error.
                                                    ' I might need to add a separate "clear error" method                                                        
    Spi.ReadChain(@encoderAngle, chainedDevices)   ' Send read command so next read will have data to retrieve.
    
PRI ProcessAngle(localIndex) | parity, error, localMovingFlag
'' Parity bit is not tested or used.
'' localIndex is the ID of the device to process.
'' The device connected to the Prop's MISO pin is
'' device #0.
'' This method performs all the software filtering of the data.
       
  'Debug.str(string(13, "ProcessAngle method"))
  parity := encoderAngle[localIndex] & %1
  encoderAngle[localIndex] >>= 1
  error := encoderAngle[localIndex] & %1
  
  if error == 1 and errorFlag < LIMIT_ERROR_MESSAGES
  ' If there is an error, display the binary value of the data.
    {Debug.str(string(13, " encoder #"))
    Debug.dec(localIndex)
    Debug.str(string(" returned an error. %")) 
    Debug.bin(encoderAngle[localIndex], 31)
    Debug.bin(parity, 1)    }
    result := 1                 ' Inform calling method of error.
  encoderAngle[localIndex] >>= 1
  encoderAngle[localIndex] &= TWELVE_BITS

  ' Compute averages, check to see if value is near rollover point and
  ' check to see if the encoder is moving.
  
  bufferTotal[localIndex] -= averagingBuffer[(localIndex * SAMPLES_TO_AVERAGE_STATIC) + bufferHead]
  averagingBuffer[(localIndex * SAMPLES_TO_AVERAGE_STATIC) + bufferHead] := encoderAngle[localIndex]
  bufferTotal[localIndex] += averagingBuffer[(localIndex * SAMPLES_TO_AVERAGE_STATIC) + bufferHead]
  average[localIndex] := bufferTotal[localIndex] / SAMPLES_TO_AVERAGE_STATIC

  bufferTotalMov[localIndex] -= averagingBufferMov[(localIndex * SAMPLES_TO_AVERAGE_MOVING) + bufferHeadM]
  averagingBufferMov[(localIndex * SAMPLES_TO_AVERAGE_MOVING) + bufferHeadM] := encoderAngle[localIndex]
  bufferTotalMov[localIndex] += averagingBufferMov[(localIndex * SAMPLES_TO_AVERAGE_MOVING) + bufferHeadM]
  averageMov[localIndex] := bufferTotalMov[localIndex] / SAMPLES_TO_AVERAGE_MOVING


  if encoderAngle[localIndex] < ROLLOVER_ZONE_LOW or encoderAngle[localIndex] > ROLLOVER_ZONE_HIGH
    rolloverTotal[localIndex] -= rolloverBuffer[(localIndex * SAMPLES_TO_AVERAGE_STATIC) + bufferHead]
    rolloverTotalMov[localIndex] -= rolloverBufferMov[(localIndex * SAMPLES_TO_AVERAGE_MOVING) + bufferHeadM]
    if encoderAngle[localIndex] < ROLLOVER_ZONE
      rolloverBuffer[(localIndex * SAMPLES_TO_AVERAGE_STATIC) + bufferHead] := encoderAngle[localIndex] + TWELVE_BITS_PLUS_ONE
      rolloverBufferMov[(localIndex * SAMPLES_TO_AVERAGE_MOVING) + bufferHeadM] := encoderAngle[localIndex] + TWELVE_BITS_PLUS_ONE
    else
      rolloverBuffer[(localIndex * SAMPLES_TO_AVERAGE_STATIC) + bufferHead] := encoderAngle[localIndex]
      rolloverBufferMov[(localIndex * SAMPLES_TO_AVERAGE_MOVING) + bufferHeadM] := encoderAngle[localIndex]
    rolloverTotal[localIndex] += rolloverBuffer[(localIndex * SAMPLES_TO_AVERAGE_STATIC) + bufferHead]
    rolloverTotalMov[localIndex] += rolloverBufferMov[(localIndex * SAMPLES_TO_AVERAGE_MOVING) + bufferHeadM]  
    if rolloverState[localIndex] == OUT_OF_ZONE_EMPTY_BUFFER
      rolloverState[localIndex] := IN_ZONE_FILLING_BUFFER
      rolloverCount[localIndex] := 0
    elseif rolloverState[localIndex] == IN_ZONE_FILLING_BUFFER
      if ++rolloverCount[localIndex] == SAMPLES_TO_AVERAGE_STATIC
        rolloverState[localIndex] := IN_ZONE_USE_BUFFER
    elseif rolloverState[localIndex] == IN_ZONE_USE_BUFFER
      average[localIndex] := (rolloverTotal[localIndex]  / SAMPLES_TO_AVERAGE_STATIC) // TWELVE_BITS_PLUS_ONE
    if rolloverStateMov[localIndex] == OUT_OF_ZONE_EMPTY_BUFFER
      rolloverStateMov[localIndex] := IN_ZONE_FILLING_BUFFER
      rolloverCountMov[localIndex] := 0
    elseif rolloverStateMov[localIndex] == IN_ZONE_FILLING_BUFFER
      if ++rolloverCountMov[localIndex] == SAMPLES_TO_AVERAGE_STATIC
        rolloverStateMov[localIndex] := IN_ZONE_USE_BUFFER
    elseif rolloverStateMov[localIndex] == IN_ZONE_USE_BUFFER
      averageMov[localIndex] := (rolloverTotalMov[localIndex]  / SAMPLES_TO_AVERAGE_MOVING) // TWELVE_BITS_PLUS_ONE  
  else ' assume rollover problems have cleared out of buffer by now
    rolloverState[localIndex] := OUT_OF_ZONE_EMPTY_BUFFER 
    rolloverStateMov[localIndex] := OUT_OF_ZONE_EMPTY_BUFFER

  localMovingFlag := NOTMOVING_THIS_LOOP
   
  if encoderAngle[localIndex] < MOVING_TRIGGER and previousAngle[localIndex] > TWELVE_BITS - MOVING_TRIGGER
    if (encoderAngle[localIndex] + TWELVE_BITS) - previousAngle[localIndex] => MOVING_TRIGGER
      localMovingFlag := MOVING_THIS_LOOP
  elseif encoderAngle[localIndex] > TWELVE_BITS - MOVING_TRIGGER and previousAngle[localIndex] < MOVING_TRIGGER
    if (previousAngle[localIndex] + TWELVE_BITS) -  encoderAngle[localIndex] => MOVING_TRIGGER
      localMovingFlag := MOVING_THIS_LOOP
  elseif ||(encoderAngle[localIndex] - previousAngle[localIndex]) => MOVING_TRIGGER  
    localMovingFlag := MOVING_THIS_LOOP

  if localMovingFlag == NOTMOVING_THIS_LOOP
    if movingState[localIndex] == MOVING
      movingState[localIndex] := NOT_MOVING_FILLING_BUFFER
      notMovingCount[localIndex] := 1
      average[localIndex] := averageMov[localIndex]
    elseif movingState[localIndex] == NOT_MOVING_FILLING_BUFFER
      if ++notMovingCount[localIndex] => SAMPLES_TO_AVERAGE_STATIC
        movingState[localIndex] := NOT_MOVING_UPDATED_BUFFER
      else
        average[localIndex] := averageMov[localIndex]
  else
    movingState[localIndex] := MOVING
    average[localIndex] := averageMov[localIndex]
  
  encoderAngle7[localIndex] := average[localIndex] 
  if average[localIndex] <> previousAngle[localIndex]   '' Comment out this line to display encoder value in each loop.
    if displayMode == SCROLL_DISPLAY 
      Debug.str(string(13, " encoder["))
      Debug.dec(localIndex)
      Debug.str(string("] = "))
      Debug.dec(average[localIndex])
    else        
      DisplayTitles
      DisplayData(localIndex)
      {Debug.str(string(13, " encoder["))
      Debug.dec(localIndex)
      Debug.str(string("] = "))
      Debug.dec(average[localIndex]) }
    {SevenSegment[localIndex].Char("v")
    if encoderAngle7[localIndex] < 1000
      SevenSegment[localIndex].Char("0")
    if encoderAngle7[localIndex] < 100
      SevenSegment[localIndex].Char("0") 
    if encoderAngle7[localIndex] < 10
      SevenSegment[localIndex].Char("0") 
    SevenSegment[localIndex].Dec(encoderAngle7[localIndex])
    }
    previousAngle[localIndex] := average[localIndex]
        
PUB DisplayHeading

  Debug.Clear
  Debug.Home
  Debug.str(string(13, 13, "-------------------------------------------------"))
  Debug.str(string(13, "Magnetic Encoder AS5055 Demo Program "))
  Debug.str(string(13, "The file name is "))
  Debug.str(@fileName)
  Debug.str(string(13, "The file name of the AS5055 driver is "))
  Debug.str(Spi.GetFileName)
  Debug.str(string(13, "The SPI driver was launched in cog #"))
  Debug.dec(spiCog)
  Debug.str(string(13, "-------------------------------------------------", 13))
  waitcnt(clkfreq + cnt)
  DisplayTitles
  'Debug.ClearBelow
                        
PUB DisplayTitles | localIndex

  repeat localIndex from 0 to finalChainIndex
    Debug.Position(titleX[localIndex], titleY[localIndex])
    Debug.str(@columnHeading)
    Debug.Dec(localIndex)
    'Debug.str(string(" = "))        ' ** added
    'Debug.Dec(average[localIndex])  ' ** added 
    
PUB DisplayData(localIndex) 

  Debug.Position(displayX[localIndex], displayY[localIndex])
  if average[localIndex] < 1000  ' space pad
    Debug.Char(32)
  if average[localIndex] < 100  
    Debug.Char(32)
  if average[localIndex] < 10  
    Debug.Char(32)        
  'Debug.Char(13)
  Debug.Dec(encoderAngle[localIndex])
  Debug.Char(32)
  Debug.Char(32)
  Debug.Char(32)
  Debug.Position(counterLocX, counterLocY)
  Debug.Dec(readingsCounter)
  
PUB Pause(ms) | t

'' Delay program in milliseconds

  if (ms < 1)                                                   ' delay must be > 0
    return
  else
    t := cnt - 1792                                             ' sync with system counter
    repeat ms                                                   ' run delay
      waitcnt(t += MS_001)
    
DAT

columnHeading           byte "Device #", 0
buffer                  byte $AA[1000]
buffer1                 byte 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15
                        byte 16, 17, 18, 19, 20, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31
                        byte 32, 64, 128, 255
buffer2                 long 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15
                        long 16, 17, 18, 19, 20, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31
                        long 32, 64, 128, 255, 256, 511, 512, 1023, 1024, 2047, 2048, 4095, 4096